
/* 
 * Process GRASS format LULC and DEM data to fix up holes (zeroes) in the 
 * data, and assign sensible color values to the various terrain features. 
 *
 * Program is called with input variables 
 *   a.out <DEM data input file> <LULC data input file> <output file name>
 * and reads in LULC color map from file LULC.colors
 * 
 * Program outputs
 *   1) where the LULC and/or DEM values are 0 (to stdout)  
 *   2) A PPM file showing only these values
 *   3) A PPM file of the whole scene
 *   4) Corrected GRASS files of DEM and LULC data in binary format
 *
*/


#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>

#define NumCategories 100
#define NumColors     3

int   color[NumCategories][NumColors];  /* color map for LULC categories */


main(int argc, char *argv[])
{
   FILE *InputDEM;
   FILE *InputLULC;
   FILE *ColorMap;
   int   OutputDEM;
   int   OutputLULC;
   int   OutputFile;
   int   BogusDataFile;
   char  Header[256];
   char  filename[256];
   char  line[4096];
   long  North = 0;
   long  South = 0;
   long  East = 0;
   long  West = 0;
   long  M = 0;
   long  N = 0;

   int  *DEM_Buffer;
   int  *Bogus_DEM_Buffer;
   char *LULC_Buffer;
   char *Bogus_LULC_Buffer;
   char *Color_Buffer;
   long  DEM_Buffer_Size;
   long  LULC_Buffer_Size;
   long  Color_Buffer_Size;
   int   i;
   int   j;
   int   Index;
   int   category;
   int   red, green, blue;

   long  DEM;
   long  LULC;
   
   if (argc < 3)
   {
      printf("usage: <GRASS DEM datafile> <GRASS LULC datafile> <output \
file name>\n");
      exit(1);
   }

   /* Open input file of DEM data */
   if ((InputDEM = fopen(argv[1], "r")) == NULL)
   {
      printf("Error -- cannot open %s\n", argv[1]);
      exit(1);
   }

   fscanf(InputDEM, "north: %ld%*c", &North);
   fscanf(InputDEM, "south: %ld%*c", &South);
   fscanf(InputDEM, "east: %ld%*c", &East);
   fscanf(InputDEM, "west: %ld%*c", &West);
   fscanf(InputDEM, "rows: %ld%*c", &M);
   fscanf(InputDEM, "cols: %ld%*c", &N);

   /* Open input file of LULC data */
   if ((InputLULC = fopen(argv[2], "r")) == NULL)
   {
      printf("Error -- cannot open %s\n", argv[2]);
      exit(1);
   }

   fscanf(InputLULC, "north: %*ld%*c");
   fscanf(InputLULC, "south: %*ld%*c");
   fscanf(InputLULC, "east: %*ld%*c");
   fscanf(InputLULC, "west: %*ld%*c");
   fscanf(InputLULC, "rows: %*ld%*c");
   fscanf(InputLULC, "cols: %*ld%*c");

   printf("(North, South, east, West) = (%ld, %ld, %ld, %ld)\n",
          North, South, East, West);
   printf("(row, col) = (%ld, %ld)\n", M, N);

   /* Open output file for binary corrected GRASS format DEM data */
   sprintf(filename, "%s.BINARY", argv[1]);
   if ((OutputDEM = open(filename, O_WRONLY | O_CREAT | O_TRUNC,
                          S_IRUSR |  S_IWUSR | S_IRGRP)) == -1)
   {
      printf("Error -- cannot open %s\n", filename);
      exit(1);
   }
   sprintf(Header, 
      "north: %ld\nsouth: %ld\neast: %ld\nwest: %ld\nrows: %ld\ncols: %ld\n\0",
       North, South, East, West, M, N);
   write(OutputDEM, Header, strlen(Header));

   /* Open output file for binary corrected GRASS format LULC data */
   sprintf(filename, "%s.BINARY", argv[2]);
   if ((OutputLULC = open(filename, O_WRONLY | O_CREAT | O_TRUNC,
                          S_IRUSR |  S_IWUSR | S_IRGRP)) == -1)
   {
      printf("Error -- cannot open %s\n", filename);
      exit(1);
   }
   sprintf(Header, 
      "north: %ld\nsouth: %ld\neast: %ld\nwest: %ld\nrows: %ld\ncols: %ld\n\0",
       North, South, East, West, M, N);
   write(OutputLULC, Header, strlen(Header));
 

   /* Set up data arrays */

   DEM_Buffer_Size = N * sizeof(int);
   LULC_Buffer_Size = N * sizeof(char);
   Color_Buffer_Size = 3 * N * sizeof(char);

   if ((DEM_Buffer = (int *) malloc(DEM_Buffer_Size)) == NULL)
   {
      printf("Out of memory!\n");
      exit(1);
   }
   if ((Bogus_DEM_Buffer = (int *) malloc(DEM_Buffer_Size)) == NULL)
   {
      printf("Out of memory!\n");
      exit(1);
   }

   if ((LULC_Buffer = (char *) malloc(LULC_Buffer_Size)) == NULL)
   {
      printf("Out of memory!\n");
      exit(1);
   }
   if ((Color_Buffer = (char *) malloc(Color_Buffer_Size)) == NULL)
   {
      printf("Out of memory!\n");
      exit(1);
   }
   if ((Bogus_LULC_Buffer = (char *) malloc(Color_Buffer_Size)) == NULL)
   {
      printf("Out of memory!\n");
      exit(1);
   }


   /* Open output PPM image file for processed data */
   sprintf(filename, "%s.ppm", argv[3]);
   if ((OutputFile = open(filename, O_WRONLY | O_CREAT | O_TRUNC,
                          S_IRUSR |  S_IWUSR | S_IRGRP)) == -1)
   {
      printf("Error -- cannot open %s\n", filename);
      exit(1);
   }

   /* Open output PPM image file for bogus data */
   sprintf(filename, "bogus-%s.ppm", argv[3]);
   if ((BogusDataFile = open(filename, O_WRONLY | O_CREAT | O_TRUNC,
                             S_IRUSR |  S_IWUSR | S_IRGRP)) == -1)
   {
      printf("Error -- cannot open %s\n", filename);
      exit(1);
   }

   sprintf(Header, "P6\n%d %d\n255\n\0", N,  M);

   write(OutputFile, Header, strlen(Header));
   write(BogusDataFile, Header, strlen(Header));

   for (i = 0; i < NumCategories; i++)
      for (j = 0; j < NumColors; j++)
	 color[i][j] = (unsigned char) 0;

   printf("reading in the LULC color map ....\n");
   fflush(stdout);

   /* Read in the LULC color map */
   if ((ColorMap = fopen("LULC.color.map", "r")) == NULL)
   {
      printf("Error -- cannot open file LULC.color\n");
      exit(1);
   }

   while (fgets(line, 4096, ColorMap) != NULL) 
   {
      printf("%s", line);
      fflush(stdout);

      /* screen for lines commented with a # */
      if (strncmp("#", line, 1)) 
      {
	 sscanf(line, "%d %d %d %d", &category, &red, &green, &blue);
	 color[category][0] = red;
 	 color[category][1] = green;
	 color[category][2] = blue;
	 printf("%d  %d %d %d\n", category, 
	        color[category][0], color[category][1], color[category][1]);
	 fflush(stdout);
      }
   }


   printf("Reading in and processing the DEM and LULC data ....\n");
   fflush(stdout);


   /* Read in and process the DEM and LULC data */
   for (i = 0; i < M; i++)
   {
      for (j = 0, Index = 0; j < N; j++, Index += 3)
      {
         fscanf(InputDEM, "%ld%*c", &DEM);
         fscanf(InputLULC, "%ld%*c", &LULC);
         
         DEM_Buffer[j] = DEM;

	 /* Currently this serves no purpose */
         Bogus_DEM_Buffer[j] = DEM;

	 if (LULC > 0)
         {
            /* Assign colors using LULC color map */           
	    Color_Buffer[Index]     = (char) color[LULC][0];
            Color_Buffer[Index + 1] = (char) color[LULC][1];
            Color_Buffer[Index + 2] = (char) color[LULC][2];
         
	    if (DEM == 0) /* Most likely water */
            {
               /* Make pixels bright blue (water?) */           
               Bogus_LULC_Buffer[Index]     = (char) 0;
               Bogus_LULC_Buffer[Index + 1] = (char) 100;
               Bogus_LULC_Buffer[Index + 2] = (char) 255;

               printf("(i, j, DEM, LULC) = (%d, %d, %ld, %ld)\n", 
			i, j, DEM, LULC);
            }
	    else
            {
               /* Make pixels white */           
               Bogus_LULC_Buffer[Index]     = (char) 255;
               Bogus_LULC_Buffer[Index + 1] = (char) 255;
               Bogus_LULC_Buffer[Index + 2] = (char) 255;
            }
         }
         else /* bogus LULC value */
         {
            printf("(i, j, DEM, LULC) = (%d, %d, %ld, %ld)\n", i, j, DEM, LULC);

	    /* Correct the problem by assigning a neighboring value, 
	       which is just the value from the previous row */
	    LULC = LULC_Buffer[j];
            printf("Assigning new LULC value %ld\n", LULC);
            
            /* Assign colors using LULC color map */           
	    Color_Buffer[Index]     = (char) color[LULC][0];
            Color_Buffer[Index + 1] = (char) color[LULC][1];
            Color_Buffer[Index + 2] = (char) color[LULC][2];
         
	    if (DEM == 0)  /* Could be water? */
            {
               /* Make pixels dark blue */           
               Bogus_LULC_Buffer[Index]     = (char) 0;
               Bogus_LULC_Buffer[Index + 1] = (char) 0;
               Bogus_LULC_Buffer[Index + 2] = (char) 100;
            }
	    else 
            {
               /* Make pixels black */           
	       Bogus_LULC_Buffer[Index]     = (char) 0;
               Bogus_LULC_Buffer[Index + 1] = (char) 0;
               Bogus_LULC_Buffer[Index + 2] = (char) 0;
            }
         }

	 /* Corrected LULC value */
	 LULC_Buffer[j] = LULC;
      }

      /* Print out row of corrected LULC and DEM data */
      write(OutputDEM, DEM_Buffer, DEM_Buffer_Size);
      write(OutputLULC, LULC_Buffer, LULC_Buffer_Size);

      /* Print out row of LULC color PPM file */
      write(OutputFile, Color_Buffer, Color_Buffer_Size);

      /* Print out row showing bogus LULC data to a PPM file */
      write(BogusDataFile, Bogus_LULC_Buffer, Color_Buffer_Size);
   }

   fclose(InputDEM);
   fclose(InputLULC);

   close(OutputDEM);
   close(OutputLULC);
   close(OutputFile);
   close(BogusDataFile);
}



